@@ -113,11 +113,12 @@ class RubyLexer
   # Last token read via next_token.
   attr_accessor :token
 
-  attr_writer :comments
+  # Last comment lexed, or nil
+  attr_accessor :comment
 
   def initialize _ = nil
     @lex_state = nil # remove one warning under $DEBUG
-    self.lex_state = EXPR_NONE
+    @lex_state = EXPR_NONE
 
     self.cond   = RubyParserStuff::StackState.new(:cond, $DEBUG)
     self.cmdarg = RubyParserStuff::StackState.new(:cmdarg, $DEBUG)
@@ -134,16 +135,6 @@ class RubyLexer
     is_after_operator? ? EXPR_ARG : EXPR_BEG
   end
 
-  def ignore_body_comments
-    @comments.clear
-  end
-
-  def comments # TODO: remove this... maybe comment_string + attr_accessor
-    c = @comments.join
-    @comments.clear
-    c
-  end
-
   def debug n
     raise "debug #{n}"
   end
@@ -258,19 +249,22 @@ class RubyLexer
   end
 
   def process_begin text
-    @comments << matched
+    self.comment ||= +""
+    self.comment << matched
 
     unless scan(/.*?\n=end( |\t|\f)*[^\n]*(\n|\z)/m) then
-      @comments.clear
+      self.comment = nil
       rb_compile_error("embedded document meets end of file")
     end
 
-    @comments << matched
+    self.comment << matched
     self.lineno += matched.count("\n") # HACK?
 
     nil # TODO
   end
 
+  # TODO: make all tXXXX terminals include lexer.lineno ... enforce it somehow?
+
   def process_brace_close text
     case matched
     when "}" then
@@ -353,8 +347,14 @@ class RubyLexer
     end
   end
 
-  def process_dots text
-    tokens = ruby27plus? && is_beg? ? BTOKENS : TOKENS
+  def process_dots text # parse32.y:10216
+    is_beg = self.is_beg?
+    self.lex_state = EXPR_BEG
+
+    return result EXPR_ENDARG, :tBDOT3, text if
+      parser.in_argdef && text == "..." # TODO: version check?
+
+    tokens = ruby27plus? && is_beg ? BTOKENS : TOKENS
 
     result EXPR_BEG, tokens[text], text
   end
@@ -439,7 +439,8 @@ class RubyLexer
 
       while scan(/\s*\#.*(\n+|\z)/) do
         self.lineno += matched.count "\n"
-        @comments << matched.gsub(/^ +#/, "#").gsub(/^ +$/, "")
+        self.comment ||= +""
+        self.comment << matched.gsub(/^ +#/, "#").gsub(/^ +$/, "")
       end
 
       return nil if end_of_stream?
@@ -687,7 +688,7 @@ class RubyLexer
       return process_token_keyword keyword if keyword
     end
 
-    # matching: compare/parse30.y:9039
+    # matching: compare/parse32.y:9031
     state = if lex_state =~ EXPR_BEG_ANY|EXPR_ARG_ANY|EXPR_DOT then
               cmd_state ? EXPR_CMDARG : EXPR_ARG
             elsif lex_state =~ EXPR_FNAME then
@@ -709,7 +710,7 @@ class RubyLexer
   end
 
   def process_token_keyword keyword
-    # matching MIDDLE of parse_ident in compare/parse23.y:8046
+    # matching MIDDLE of parse_ident in compare/parse32.y:9695
     state = lex_state
 
     return result(EXPR_ENDFN, keyword.id0, token) if lex_state =~ EXPR_FNAME
@@ -718,7 +719,7 @@ class RubyLexer
     self.command_start = true if lex_state =~ EXPR_BEG
 
     case
-    when keyword.id0 == :kDO then # parse26.y line 7591
+    when keyword.id0 == :kDO then # parse32.y line 9712
       case
       when lambda_beginning? then
         self.lpar_beg = nil # lambda_beginning? == FALSE in the body of "-> do ... end"
@@ -760,7 +761,7 @@ class RubyLexer
     self.lineno        = 1
     self.brace_nest    = 0
     self.command_start = true
-    self.comments      = []
+    self.comment       = nil
     self.lex_state     = EXPR_NONE
     self.lex_strterm   = nil
     self.lpar_beg      = nil
@@ -890,7 +891,7 @@ class RubyLexer
     end
 
     def inspect
-      return "Value(0)" if n.zero? # HACK?
+      return "EXPR_NONE" if n.zero? # HACK?
 
       names.map { |v, k| k if self =~ v }.
         compact.
