@@ -18,7 +18,7 @@ class Sexp
   end
 
   ##
-  # Returns the maximum line number of the children of self.
+  # Returns the minimum line number of the children of self.
 
   def line_min
     @line_min ||= [self.deep_each.map(&:line).min, self.line].compact.min
@@ -30,11 +30,11 @@ class Sexp
 end
 
 module RubyParserStuff
-  VERSION = "3.20.0"
+  VERSION = "3.21.1"
 
-  attr_accessor :lexer, :in_def, :in_single, :file
+  attr_accessor :lexer, :in_def, :in_single, :file, :in_argdef
   attr_accessor :in_kwarg
-  attr_reader :env, :comments
+  attr_reader :env
 
   ##
   # Canonicalize conditionals. Eg:
@@ -92,6 +92,7 @@ module RubyParserStuff
     [k, true]
   }.to_h
 
+  # TODO: remove
   has_enc = "".respond_to? :encoding
 
   # This is in sorted order of occurrence according to
@@ -122,9 +123,9 @@ module RubyParserStuff
     self.lexer = RubyLexer.new v && v.to_i
     self.lexer.parser = self
     self.in_kwarg = false
+    self.in_argdef = false
 
     @env = RubyParserStuff::Environment.new
-    @comments = []
 
     @canonicalize_conditions = true
 
@@ -153,6 +154,7 @@ module RubyParserStuff
       result.line lexer.lineno
     else
       result.line ss.first.line
+      result.line_max = ss.first.line_max
     end
 
     args.each do |arg|
@@ -330,9 +332,8 @@ module RubyParserStuff
     end
 
     args.each do |arg|
-      if arg.instance_of? Array and arg.size == 2 and arg.last.is_a? Numeric then
-        arg = arg.first
-      end
+      # ruby 3.0+ TODO: next if arg in [String, Integer] # eg ["(", 1]
+      next if arg.class == Array && arg.map(&:class) == [String, Integer]
 
       case arg
       when Sexp then
@@ -566,7 +567,7 @@ module RubyParserStuff
 
   def handle_encoding str
     str = str.dup
-    has_enc = str.respond_to? :encoding
+    has_enc = str.respond_to? :encoding # TODO: remove
     encoding = nil
 
     header = str.each_line.first(2)
@@ -663,6 +664,7 @@ module RubyParserStuff
       end
 
       if head.size == 2 and tail.size > 1 and tail[1].sexp_type == :str then
+        head[-1] = head[-1].dup if head[-1].frozen?
         head.last << tail[1].last
         head.sexp_type = :str if head.size == 2 # HACK ?
       else
@@ -794,6 +796,7 @@ module RubyParserStuff
     case lhs.sexp_type
     when :lasgn, :iasgn, :cdecl, :cvdecl, :gasgn, :cvasgn, :attrasgn, :safe_attrasgn then
       lhs << rhs
+      lhs.line_max = rhs.line_max
     when :const then
       lhs.sexp_type = :cdecl
       lhs << rhs
@@ -823,10 +826,10 @@ module RubyParserStuff
   end
 
   def new_begin val
-    _, lineno, body, _ = val
+    (_, line), _, body, _ = val
 
     result = body ? s(:begin, body) : s(:nil)
-    result.line lineno
+    result.line line
 
     result
   end
@@ -885,12 +888,13 @@ module RubyParserStuff
     # TODO: need a test with f(&b) to produce block_pass
     # TODO: need a test with f(&b) { } to produce warning
 
-    if args
+    if args then
       if ARG_TYPES[args.sexp_type] then
         result.concat args.sexp_body
       else
         result << args
       end
+      result.line_max = args.line_max
     end
 
     # line = result.grep(Sexp).map(&:line).compact.min
@@ -926,8 +930,7 @@ module RubyParserStuff
   end
 
   def new_class val
-    # TODO: get line from class keyword
-    line, path, superclass, body = val[1], val[2], val[3], val[5]
+    (_, line, comment), path, superclass, _, body, (_, line_max) = val
 
     path = path.first if path.instance_of? Array
 
@@ -942,7 +945,8 @@ module RubyParserStuff
     end
 
     result.line = line
-    result.comments = self.comments.pop
+    result.line_max = line_max
+    result.comments = comment if comment
     result
   end
 
@@ -970,13 +974,18 @@ module RubyParserStuff
   end
 
   def new_defn val
-    _, (name, line), in_def, args, body, _ = val
+    if val.size == 4 then
+      ((_, line, comment), (name, _line, in_def)), args, body, (_, line_max) = val
+    else
+      (_, line, comment), (name, line), in_def, args, body, (_, line_max) = val
+    end
 
     body ||= s(:nil).line line
 
     args.line line
 
     result = s(:defn, name.to_sym, args).line line
+    result.line_max = line_max
 
     if body.sexp_type == :block then
       result.push(*body.sexp_body)
@@ -984,13 +993,14 @@ module RubyParserStuff
       result.push body
     end
 
-    result.comments = self.comments.pop
+    result.comments = comment if comment
 
     [result, in_def]
   end
 
   def new_endless_defn val
-    (name, line, in_def), args, _, body, _, resbody = val
+    # not available in 2.x so we don't need to check size
+    ((_, line, comment), (name, _, in_def)), args, _, body, _, resbody = val
 
     result =
       if resbody then
@@ -1005,13 +1015,15 @@ module RubyParserStuff
     local_pop in_def
     endless_method_name result
 
-    result.comments = self.comments.pop
+    result.comments = comment if comment
 
     result
   end
 
   def new_endless_defs val
-    (recv, (name, line, in_def)), args, _, body, _, resbody = val
+    # not available in 2.x so we don't need to check size
+    ((_, line, comment), recv, _, _, (name, line, in_def)), \
+      args, _, body, _, resbody = val
 
     result =
       if resbody then
@@ -1027,19 +1039,26 @@ module RubyParserStuff
     local_pop in_def
     endless_method_name result
 
-    result.comments = self.comments.pop
+    result.comments = comment if comment
 
     result
   end
 
   def new_defs val
-    _, recv, (name, line), in_def, args, body, _ = val
+    if val.size == 4 then
+      ((_, line, comment), recv, _, _, (name, line, in_def)), \
+        args, body, (_, line_max) = val
+    else
+      (_, line, comment), recv, (name, _), in_def, \
+        args, body, (_, line_max) = val
+    end
 
     body ||= s(:nil).line line
 
     args.line line
 
     result = s(:defs, recv, name.to_sym, args).line line
+    result.line_max = line_max
 
     # TODO: remove_begin
     # TODO: reduce_nodes
@@ -1050,7 +1069,7 @@ module RubyParserStuff
       result.push body
     end
 
-    result.comments = self.comments.pop
+    result.comments = comment if comment
 
     [result, in_def]
   end
@@ -1204,12 +1223,12 @@ module RubyParserStuff
   end
 
   def new_module val
-    # TODO: get line from module keyword
-    line, path, body = val[1], val[2], val[4]
+    (_, line_min, comment), path, _, body, (_, line_max) = val
 
     path = path.first if path.instance_of? Array
 
-    result = s(:module, path).line line
+    result = s(:module, path).line line_min
+    result.line_max = line_max
 
     if body then # REFACTOR?
       if body.sexp_type == :block then
@@ -1219,7 +1238,7 @@ module RubyParserStuff
       end
     end
 
-    result.comments = self.comments.pop
+    result.comments = comment if comment
     result
   end
 
@@ -1291,9 +1310,10 @@ module RubyParserStuff
   end
 
   def new_regexp val
-    (_, line), node, (options, _) = val
+    (_, line), node, (options, line_max) = val
 
     node ||= s(:str, "").line line
+    node.line_max = line_max
 
     o, k = 0, nil
     options.split(//).uniq.each do |c| # FIX: this has a better home
@@ -1361,7 +1381,7 @@ module RubyParserStuff
   end
 
   def new_sclass val
-    recv, in_def, in_single, body = val[3], val[4], val[6], val[7]
+    (_, line), _, recv, in_def, _, in_single, body, _ = val
 
     result = s(:sclass, recv)
 
@@ -1373,7 +1393,7 @@ module RubyParserStuff
       end
     end
 
-    result.line = val[2]
+    result.line = line
     self.in_def = in_def
     self.in_single = in_single
     result
@@ -1512,11 +1532,16 @@ module RubyParserStuff
     end
   end
 
+  KEEP_COMMENT_TOKENS = [:kCLASS, :kMODULE, :kDEF, :tNL]
+
   def next_token
     token = self.lexer.next_token
 
     if token and token.first != RubyLexer::EOF then
       self.last_token_type = token
+
+      self.lexer.comment = nil unless KEEP_COMMENT_TOKENS.include? token.first
+
       return token
     elsif !token
       return self.lexer.next_token
@@ -1576,7 +1601,6 @@ module RubyParserStuff
     self.in_def = false
     self.in_single = 0
     self.env.reset
-    self.comments.clear
     self.last_token_type = nil
   end
 
